/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*-
 *
 * Copyright (c) 1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * Ported from CMU/Monarch's code, nov'98 -Padma.
 * Pradeep Kyasanur, March 2003: New class for simulating MAC layer misbehavior, derives code
 * from mac-802_11.cc file
 */


#include "delay.h"
#include "connector.h"
#include "packet.h"
#include "random.h"
#include "mobilenode.h"

// #define DEBUG

#include "arp.h"
#include "ll.h"
#include "mac.h"
#include "mac-timers.h"
#include "mac-802_11.h"
#include "cmu-trace.h" 
#include "mac-802_11mb.h" 



// XXX Can't we make these macros inline methods? Otherwise why should we have
// inline methods at all??

#define CHECK_BACKOFF_TIMER()						\
{									\
	if(is_idle() && mhBackoff_.paused())				\
		mhBackoff_.resume(difs_);				\
	if(! is_idle() && mhBackoff_.busy() && ! mhBackoff_.paused())	\
		mhBackoff_.pause();					\
}

#define TRANSMIT(p, t)                                                  \
{                                                                       \
	tx_active_ = 1;                                                  \
                                                                        \
        /*                                                              \
         * If I'm transmitting without doing CS, such as when           \
         * sending an ACK, any incoming packet will be "missed"         \
         * and hence, must be discarded.                                \
         */                                                             \
        if(rx_state_ != MAC_IDLE) {                                      \
                struct hdr_mac802_11 *dh = HDR_MAC802_11(p);                  \
                                                                        \
                assert(dh->dh_fc.fc_type == MAC_Type_Control);          \
                assert(dh->dh_fc.fc_subtype == MAC_Subtype_ACK);        \
                                                                        \
                assert(pktRx_);                                          \
                struct hdr_cmn *ch = HDR_CMN(pktRx_);                    \
                                                                        \
                ch->error() = 1;        /* force packet discard */      \
        }                                                               \
                                                                        \
        /*                                                              \
         * pass the packet on the "interface" which will in turn        \
         * place the packet on the channel.                             \
         *                                                              \
         * NOTE: a handler is passed along so that the Network          \
         *       Interface can distinguish between incoming and         \
         *       outgoing packets.                                      \
         */                                                             \
        downtarget_->recv(p->copy(), this);                             \
                                                                        \
        mhSend_.start(t);                                                \
                                                                        \
	mhIF_.start(TX_Time(p));                                         \
}

#define SET_RX_STATE(x)			\
{					\
	rx_state_ = (x);		\
	CHECK_BACKOFF_TIMER();		\
}

#define SET_TX_STATE(x)				\
{						\
	tx_state_ = (x);			\
	CHECK_BACKOFF_TIMER();			\
}

/* ======================================================================
   Global Variables
   ====================================================================== */

static PHY_MIB PMIB =
	{
		DSSS_CWMin, DSSS_CWMax, DSSS_SlotTime, DSSS_CCATime,
		DSSS_RxTxTurnaroundTime, DSSS_SIFSTime, DSSS_PreambleLength,
		DSSS_PLCPHeaderLength
	};	

static MAC_MIB MMIB =
	{
		0 /* MAC_RTSThreshold */, MAC_ShortRetryLimit,
		MAC_LongRetryLimit, MAC_FragmentationThreshold,
		MAC_MaxTransmitMSDULifetime, MAC_MaxReceiveLifetime,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
	};

/* ======================================================================
   TCL Hooks for the simulator
   ====================================================================== */
static class Mac802_11mbClass : public TclClass {
public:
	Mac802_11mbClass() : TclClass("Mac/802_11mb") {}
	TclObject* create(int, const char*const*) {
		return (new Mac802_11mb(&PMIB, &MMIB));
	}
} class_mac802_11mb;


/* ======================================================================
   Mac Class Functions
   ====================================================================== */
Mac802_11mb::Mac802_11mb(PHY_MIB *p, MAC_MIB *m) : Mac(), 
						   mhIF_(this), mhNav_(this), mhRecv_(this), mhSend_(this), 
						   mhDefer_(this, p->SlotTime), mhBackoff_(this, p->SlotTime) 
{ 
	macmib_ = m; 
	phymib_ = p; 
 
	nav_ = 0.0; 

	// 3/12/03 Pradeep Kyasanur: Initialize a modified NAV variable for monitoring purposes
	modified_nav_ = 0.0;
 
	tx_state_ = rx_state_ = MAC_IDLE; 
	tx_active_ = 0; 
 
	pktRTS_ = 0; 
	pktCTRL_ = 0; 
 
	cw_ = phymib_->CWMin; 
	ssrc_ = slrc_ = 0; 
 
	sifs_ = phymib_->SIFSTime; 
	pifs_ = sifs_ + phymib_->SlotTime; 
	difs_ = sifs_ + 2*phymib_->SlotTime; 
	eifs_ = sifs_ + difs_ + DATA_Time(ETHER_ACK_LEN + 
					  phymib_->PreambleLength/8 + phymib_->PLCPHeaderLength/8); 
	tx_sifs_ = sifs_ - phymib_->RxTxTurnaroundTime; 
	tx_pifs_ = tx_sifs_ + phymib_->SlotTime; 
	tx_difs_ = tx_sifs_ + 2 * phymib_->SlotTime; 
 
	sta_seqno_ = 1; 
	cache_ = 0; 
	cache_node_count_ = 0; 
 
	/*Pradeep Kyasanur: Code added for monitoring and misbehaving handling */ 
	monitor_=0; 
	backoffSender=0; 
	backoffReceiver=0;
} 
	
	

int
Mac802_11mb::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		if (strcmp(argv[1], "log-target") == 0) {
			// NsObject replaced by Trace by Pradeep Kyasanur on 05/21
			// This allows controlling trace output better and avoid getting large
			// trace files
			logtarget_ = (Trace *) TclObject::lookup(argv[2]);
 
			if(logtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		} else if(strcmp(argv[1], "nodes") == 0) {
			if(cache_) return TCL_ERROR;
			cache_node_count_ = atoi(argv[2]);
			cache_ = new Host[cache_node_count_ + 1];
			assert(cache_);
			bzero(cache_, sizeof(Host) * (cache_node_count_+1 ));
			return TCL_OK;
		}
		

		/* Pradeep Kyasanur: The next three functions have been added to 
		 * set up properties of nodes
		 * 1. BackoffSender is a class in the receiver to compute the backoff values
		 *    to be assigned to the sender for next transmission to the receiver. This
		 *    class is enabled to assign backoff values when initialized with a pointer to
		 *    monitor_ object
		 * 2. BackoffReceiver is a class in the sender, which stores the backoff value 
		 *    assigned by the sender and controls the backoff counter with the assigned 
		 *    backoff
		 * 3. misb-target() initializes the trace object to which log data is sent
		 */
		else if(strcmp(argv[1], "set-backoff-sender") == 0) {
			if(atoi(argv[2]) > 0)
				backoffSender= new BackoffSender(this,monitor_);
			else
				backoffSender= new BackoffSender(this,0);
			return TCL_OK;; 
		} 
		else if(strcmp(argv[1], "set-backoff-receiver") == 0) { 
			backoffReceiver= new BackoffReceiver(this); 
			return TCL_OK;; 
		}else if (strcmp(argv[1], "misb-target") == 0) {  // log misbehave data
			misbtarget_ = (Trace *) TclObject::lookup(argv[2]);
			if(misbtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		} 

	}
	else if(argc == 5)
		{

			// Pradeep Kyasanur: Controls setting backoff policies
			if(strcmp(argv[1], "backoff-policy") == 0) {
				int backoff_policy=atoi(argv[2]);
				int nodeID = atoi(argv[3]);
				int MISB_PERC=atoi(argv[4]);

				mhBackoff_.set_misbehavior_percentage(MISB_PERC);
				/* Pradeep Kyasanur: Updated July 22
				 * Allowed backoff policies  
				 * 0: Normal 802.11 scheme
				 * 1: Backoff always set to MISB_PERC slots
				 * 2: Backoff always set to MISB_PERC of the assigned backoff
				 * Any number >= 3: Backoff is set to misb% of assigned backoff.
				 * However, the misb% is adaptively changed based on response to 
				 * misbehavior. The amount to decrement by is set to MISB_PERC 
				 * (Adaptive misbehavior is discussed in journal paper)
				 */
				backoffReceiver->dec = 0; // by default there is no adaptive misbehavior
				if(backoff_policy >=0)  // currently 4 policies are allowed 
					{

						if(backoff_policy >= 3) {
							mhBackoff_.set_backoff_policy(2,nodeID);
							backoffReceiver->misb = 100;
							backoffReceiver->backoffTimer = &mhBackoff_;
						        backoffReceiver->dec = MISB_PERC;
							mhBackoff_.set_misbehavior_percentage(backoffReceiver->misb);
						} else {
							
							mhBackoff_.set_backoff_policy(backoff_policy,nodeID);
						}
						return TCL_OK;
					}
				return TCL_ERROR;
			}
			/* Pradeep Kyasanur: 
			 * Monitor is a class that monitors the channel, and identifies any
			 * misbehaving nodes. This is generally enabled at the receiver along
			 * with BackoffSender class. In particular, once a sender is assigned a 
			 * backoff, Monitor tracks the number of slots waited for by the sender,
			 * before the next packet transmission. To enable accurate tracking, there
			 * is code dispersed throughout this file to track the values when various
			 * operations such as updating of NAV, loss of CTS/ACK, etc. take place 
			 */
			else if(strcmp(argv[1],"set-monitor")==0) {
			
				 
				double jitter=atof(argv[2]);
				int thresh=atoi(argv[3]);
				int k=atoi(argv[4]);
				monitor_= new Monitor(this,jitter,thresh,k);
				return TCL_OK;; 	
			}
		}

	return Mac::command(argc, argv);
}

/* ======================================================================
   Debugging Routines
   ====================================================================== */
void
Mac802_11mb::trace_pkt(Packet *p) {
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac802_11* dh = HDR_MAC802_11(p);
	u_int16_t *t = (u_int16_t*) &dh->dh_fc;

	fprintf(stderr, "\t[ %2x %2x %2x %2x ] %x %s %d\n",
		*t, dh->dh_duration,
		ETHER_ADDR(dh->dh_da), ETHER_ADDR(dh->dh_sa),
		index_, packet_info.name(ch->ptype()), ch->size());
}

void
Mac802_11mb::dump(char *fname)
{
	fprintf(stderr,
		"\n%s --- (INDEX: %d, time: %2.9f)\n",
		fname, index_, Scheduler::instance().clock());

	fprintf(stderr,
		"\ttx_state_: %x, rx_state_: %x, nav: %2.9f, idle: %d\n",
		tx_state_, rx_state_, nav_, is_idle());

	fprintf(stderr,
		"\tpktTx_: %x, pktRx_: %x, pktRTS_: %x, pktCTRL_: %x, callback: %x\n",
		(int) pktTx_, (int) pktRx_, (int) pktRTS_,
		(int) pktCTRL_, (int) callback_);

	fprintf(stderr,
		"\tDefer: %d, Backoff: %d (%d), Recv: %d, Timer: %d Nav: %d\n",
		mhDefer_.busy(), mhBackoff_.busy(), mhBackoff_.paused(),
		mhRecv_.busy(), mhSend_.busy(), mhNav_.busy());
	fprintf(stderr,
		"\tBackoff Expire: %f\n",
		mhBackoff_.expire());
}


/* ======================================================================
   Packet Headers Routines
   ====================================================================== */
inline int
Mac802_11mb::hdr_dst(char* hdr, int dst )
{
	struct hdr_mac802_11 *dh = (struct hdr_mac802_11*) hdr;
	//dst = (u_int32_t)(dst);

	if(dst > -2)
		STORE4BYTE(&dst, (dh->dh_da));

	return ETHER_ADDR(dh->dh_da);
}

inline int 
Mac802_11mb::hdr_src(char* hdr, int src )
{
	struct hdr_mac802_11 *dh = (struct hdr_mac802_11*) hdr;
	if(src > -2)
		STORE4BYTE(&src, (dh->dh_sa));
	return ETHER_ADDR(dh->dh_sa);
}

inline int 
Mac802_11mb::hdr_type(char* hdr, u_int16_t type)
{
	struct hdr_mac802_11 *dh = (struct hdr_mac802_11*) hdr;
	if(type)
		STORE2BYTE(&type,(dh->dh_body));
	return GET2BYTE(dh->dh_body);
}


/* ======================================================================
   Misc Routines
   ====================================================================== */


/* Pradeep Kyasanur: 
 *Some debugging code has been commented out
 */
inline int
Mac802_11mb::is_idle()
{
	if(rx_state_ != MAC_IDLE)
		{
			//	if(getNodeId()==3)
			//	printf("Channel found to be busy by rx state\n");
			return 0; 
		}
	if(tx_state_ != MAC_IDLE)
		{
			//	if(getNodeId()==3)
			//	printf("Channel found to be busy  by tx_state\n");
			return 0; 
		}
	if(nav_ > Scheduler::instance().clock())
		{
			//	if(getNodeId()==3)
			//	printf("Channel found to be busy  by nav_ = %f\n",nav_/DSSS_SlotTime); 
			return 0; 
		}
	/* Pradeep Kyasanur: 
	 * channelBusy() is a new function added to wireless_phy.cc.
	 * The channel model in original ns distribution assumes that channel status 
	 * does not change for every slot. With this assumption, our protocol will perform
	 * very well (as chance for misdiagnosis reduces), but we won't have a realistic
	 * evaluation. Hence, the wireless physical layer model has been updated to 
	 * reevaluate channel status at every slot (e.g., there may be fading). This 
	 * modification is useful only when shadowing (or other random) channel model is used. If
	 * two-rayground model is used, then there are no variations in received single strength
	 * with time, and so slot by slot evaluation is unnecessary.
	 * As a consequence of this modification, we really need to evaluate the channel status
	 * every slot to see if the channel continues to be idle (busy)
	 */
	if(netif_->channelBusy()) 
		{ 
			//	if(getNodeId()==3)
			//	printf("Channel found to be busy  \n"); 
			return 0; 
		} 
	
	return 1;
}



/* 3/12/03 Pradeep Kyasanur: The is_idle() function is extended to enable the IdleCounter to count
 * channel to be busy only under following conditions
 * 1. The node is sending some data
 * 2. The node is receiving data at strength above the receive threshold
 * 3. Overheard RTS/CTS/DATA/ACK have set the nav
 * For the purposes of the implementation, we use a new nav variable and specially set it to allow
 * interoperability with unmodified version. 
 * This modification is for reducing misdiagnosis probability, as described in journal paper.
 */


inline int
Mac802_11mb::is_idle_modified()
{
	if(tx_state_ != MAC_IDLE)
		{
			return 0; 
		}
	
	if(rx_state_ != MAC_IDLE) 
		{  
			// only if the packet is being received without errors, set the channel to be busy
			// this is because, ns physical layer is a hack and sets the error bit in packet
			// header if packet is received with power between RX and CS thresholds

			assert(pktRx_);                                       
			struct hdr_cmn *ch = HDR_CMN(pktRx_);  
			if( ch->error() == 0 ) 
				return 0; 
		}
	
	if(modified_nav_ > Scheduler::instance().clock())
		{
			return 0; 
		}
	
	return 1;
}


inline void Mac802_11mb :: set_modified_nav(u_int16_t us) {
	double now = Scheduler::instance().clock();
	double t = us * 1e-6;

	if((now + t) > modified_nav_) {
		modified_nav_ = now + t;
		
	}
}
 


// -------------------------------------------------------------------------------

void
Mac802_11mb::discard(Packet *p, const char* why)
{
	hdr_mac802_11* mh = HDR_MAC802_11(p);
	hdr_cmn *ch = HDR_CMN(p);

#if 0
	/* old logic 8/8/98 -dam */
	/*
	 * If received below the RXThreshold, then just free.
	 */
	if(p->txinfo_.Pr < p->txinfo_.ant.RXThresh) {
		Packet::free(p);
		//p = 0;
		return;
	}
#endif // 0

	/* if the rcvd pkt contains errors, a real MAC layer couldn't
	   necessarily read any data from it, so we just toss it now */
	if(ch->error() != 0) {
		Packet::free(p);
		//p = 0;
		return;
	}

	switch(mh->dh_fc.fc_type) {
	case MAC_Type_Management:
		drop(p, why);
		return;
	case MAC_Type_Control:
		switch(mh->dh_fc.fc_subtype) {
		case MAC_Subtype_RTS:
			if((u_int32_t)ETHER_ADDR(mh->dh_sa) == \
			   (u_int32_t)index_) {
				drop(p, why);
				return;
			}
			/* fall through - if necessary */
		case MAC_Subtype_CTS:
		case MAC_Subtype_ACK:
			if((u_int32_t)ETHER_ADDR(mh->dh_da) == \
			   (u_int32_t)index_) {
				drop(p, why);
				return;
			}
			break;
		default:
			fprintf(stderr, "invalid MAC Control subtype\n");
			exit(1);
		}
		break;
	case MAC_Type_Data:
		switch(mh->dh_fc.fc_subtype) {
		case MAC_Subtype_Data:
			if((u_int32_t)ETHER_ADDR(mh->dh_da) == \
			   (u_int32_t)index_ ||
			   (u_int32_t)ETHER_ADDR(mh->dh_sa) == \
			   (u_int32_t)index_ ||
			   (u_int32_t)ETHER_ADDR(mh->dh_da) == MAC_BROADCAST) {
				drop(p);
				return;
			}
			break;
		default:
			fprintf(stderr, "invalid MAC Data subtype\n");
			exit(1);
		}
		break;
	default:
		fprintf(stderr, "invalid MAC type (%x)\n", mh->dh_fc.fc_type);
		trace_pkt(p);
		exit(1);
	}
	Packet::free(p);
}

void
Mac802_11mb::capture(Packet *p)
{
	/*
	 * Update the NAV so that this does not screw
	 * up carrier sense.
	 */
	set_nav(usec(eifs_ + TX_Time(p)));
	Packet::free(p);
}

void
Mac802_11mb::collision(Packet *p)
{
	switch(rx_state_) {
	case MAC_RECV:
		SET_RX_STATE(MAC_COLL);
		/* fall through */
	case MAC_COLL:
		assert(pktRx_);
		assert(mhRecv_.busy());
		/*
		 *  Since a collision has occurred, figure out
		 *  which packet that caused the collision will
		 *  "last" the longest.  Make this packet,
		 *  pktRx_ and reset the Recv Timer if necessary.
		 */
		if(TX_Time(p) > mhRecv_.expire()) {
			mhRecv_.stop();
			discard(pktRx_, DROP_MAC_COLLISION);
			pktRx_ = p;
			mhRecv_.start(TX_Time(pktRx_));
		}
		else {
			discard(p, DROP_MAC_COLLISION);
		}
		break;
	default:
		assert(0);
	}
}

void
Mac802_11mb::tx_resume()
{
	assert(mhSend_.busy() == 0);
	assert(mhDefer_.busy() == 0);

	if(pktCTRL_) {
		/*
		 *  Need to send a CTS or ACK.
		 */
		mhDefer_.start(sifs_);
	} else if(pktRTS_) {
		if(mhBackoff_.busy() == 0)
			mhDefer_.start(difs_);
	} else if(pktTx_) {
		if(mhBackoff_.busy() == 0)
			mhDefer_.start(difs_);
	} else if(callback_) {
		Handler *h = callback_;
		callback_ = 0;
		h->handle((Event*) 0);
	}
	SET_TX_STATE(MAC_IDLE);
}

void
Mac802_11mb::rx_resume()
{
	assert(pktRx_ == 0);
	assert(mhRecv_.busy() == 0);
	SET_RX_STATE(MAC_IDLE);
}


/* ======================================================================
   Timer Handler Routines
   ====================================================================== */
void
Mac802_11mb::backoffHandler()
{
	if(pktCTRL_) {
		assert(mhSend_.busy() || mhDefer_.busy());
		return;
	}

	if(check_pktRTS() == 0)
		return;

	if(check_pktTx() == 0)
		return;
}

void
Mac802_11mb::deferHandler()
{
	assert(pktCTRL_ || pktRTS_ || pktTx_);

	if(check_pktCTRL() == 0)
		return;

	assert(mhBackoff_.busy() == 0);

	if(check_pktRTS() == 0)
		return;

	if(check_pktTx() == 0)
		return;
}

void
Mac802_11mb::navHandler()
{
	if(is_idle() && mhBackoff_.paused())
		mhBackoff_.resume(difs_);
}

void
Mac802_11mb::recvHandler()
{
	recv_timer();
}

void
Mac802_11mb::sendHandler()
{
	send_timer();
}


void
Mac802_11mb::txHandler()
{
	tx_active_ = 0;
}


/* ======================================================================
   The "real" Timer Handler Routines
   ====================================================================== */
void
Mac802_11mb::send_timer()
{
	switch(tx_state_) {
		/*
		 * Sent a RTS, but did not receive a CTS.
		 */
	case MAC_RTS:
		RetransmitRTS();
		break;
		/*
		 * Sent a CTS, but did not receive a DATA packet.
		 */
	case MAC_CTS:
		assert(pktCTRL_); 
		 
		// In this case, have to restore the backoff value in the Monitor 
		if(monitor_) 
			monitor_->restoreBackoff(pktCTRL_); 

		Packet::free(pktCTRL_); pktCTRL_ = 0;
		break;
		/*
		 * Sent DATA, but did not receive an ACK packet.
		 */
	case MAC_SEND:
		RetransmitDATA();
		break;
		/*
		 * Sent an ACK, and now ready to resume transmission.
		 */
	case MAC_ACK:
		assert(pktCTRL_);
		Packet::free(pktCTRL_); pktCTRL_ = 0;
		break;
	case MAC_IDLE:
		break;
	default:
		assert(0);
	}
	tx_resume();
}


/* ======================================================================
   Outgoing Packet Routines
   ====================================================================== */
int
Mac802_11mb::check_pktCTRL()
{
	struct hdr_mac802_11 *mh;
	double timeout;

	if(pktCTRL_ == 0)
		return -1;
	if(tx_state_ == MAC_CTS || tx_state_ == MAC_ACK)
		return -1;

	mh = HDR_MAC802_11(pktCTRL_);
							  
	switch(mh->dh_fc.fc_subtype) {
		/*
		 *  If the medium is not IDLE, don't send the CTS.
		 */
	case MAC_Subtype_CTS:
		if(!is_idle()) { 
 
			//Pradeep Kyasanur: In this case we must not update the monitor value 
			if(monitor_) 
				monitor_->restoreBackoff(pktCTRL_); 

			discard(pktCTRL_, DROP_MAC_BUSY); pktCTRL_ = 0;
			return 0;
		}
		SET_TX_STATE(MAC_CTS);
		

		
		timeout = CTS_Time + sifs_ + DATA_Time(700); // 2000 is assumed to be max packet size
		break;
		/*
		 * IEEE 802.11 specs, section 9.2.8
		 * Acknowledments are sent after an SIFS, without regard to
		 * the busy/idle state of the medium.
		 */
	case MAC_Subtype_ACK:
		SET_TX_STATE(MAC_ACK);
		timeout = ACK_Time;
		break;
	default:
		fprintf(stderr, "check_pktCTRL:Invalid MAC Control subtype\n");
		exit(1);
	} 
 
	/*Pradeep Kyasanur: code for Monitoring */ 
	if(tx_state_ == MAC_ACK && monitor_) 
		monitor_->updateLastCount(pktCTRL_); 
	 
	 

        TRANSMIT(pktCTRL_, timeout);
	return 0;
}

int
Mac802_11mb::check_pktRTS()
{
	struct hdr_mac802_11 *mh;
	double timeout;

	assert(mhBackoff_.busy() == 0);

	if(pktRTS_ == 0)
 		return -1;
	//struct hdr_cmn *ch = HDR_CMN(pktRTS_);
	mh = HDR_MAC802_11(pktRTS_);

 	switch(mh->dh_fc.fc_subtype) {
	case MAC_Subtype_RTS:
		if(! is_idle()) {
			inc_cw(); 
		 
			/*Pradeep Kyasanur: code for handling backoff receiving */ 
			int backoff; 
			if(backoffReceiver) 
				backoff= backoffReceiver->getBackoff((u_int32_t)ETHER_ADDR(mh->dh_da)); 
			else 
				backoff = (Random::random() % cw_) + 1;
			mhBackoff_.start(backoff, is_idle()); 

			return 0;
		}
		SET_TX_STATE(MAC_RTS);
		timeout = CTSTimeout;
		break;
	default:
		fprintf(stderr, "check_pktRTS:Invalid MAC Control subtype\n");
		exit(1);
	}
	// Pradeep Kyasanur: Code added here for sending the attempt number in RTS packets
	struct rts_frame *rf = (struct rts_frame*)pktRTS_->access(hdr_mac::offset_);


     

	u_int32_t dst = ETHER_ADDR(mh->dh_da);

	if(backoffReceiver)
		rf->rf_attempt=(u_int16_t)backoffReceiver->getAttemptNumber(dst);


        TRANSMIT(pktRTS_, timeout);
	return 0;
}

int
Mac802_11mb::check_pktTx()
{
	struct hdr_mac802_11 *mh;
	double timeout;
	
	assert(mhBackoff_.busy() == 0);

	if(pktTx_ == 0)
		return -1;

	mh = HDR_MAC802_11(pktTx_);
       	int len = HDR_CMN(pktTx_)->size();

	switch(mh->dh_fc.fc_subtype) {
	case MAC_Subtype_Data:
		if(! is_idle()) {
			sendRTS(ETHER_ADDR(mh->dh_da));
			inc_cw(); 
 
		
			// Pradeep Kyasanur:
			int backoff; 
			if(backoffReceiver) 
				backoff= backoffReceiver->getBackoff((u_int32_t)ETHER_ADDR(mh->dh_da)); 
			else 
				backoff = (Random::random() % cw_) + 1; 
			mhBackoff_.start(backoff, is_idle()); 

			//			mhBackoff_.start(cw_, is_idle());
			return 0;
		}
		SET_TX_STATE(MAC_SEND);
		if((u_int32_t)ETHER_ADDR(mh->dh_da) != MAC_BROADCAST)
			//timeout = ACKTimeout(netif_->txtime(pktTx_))+5;
			// why 10 ? buggy
			//timeout = ACKTimeout(len) + 10;
			timeout = ACKTimeout(len);
		else
			timeout = TX_Time(pktTx_);
		break;
	default:
		fprintf(stderr, "check_pktTx:Invalid MAC Control subtype\n");
		//printf("pktRTS:%x, pktCTS/ACK:%x, pktTx:%x\n",pktRTS_, pktCTRL_,pktTx_);
		exit(1);
	}
        TRANSMIT(pktTx_, timeout);
	return 0;
}

/*
 * Low-level transmit functions that actually place the packet onto
 * the channel.
 */
void
Mac802_11mb::sendRTS(int dst)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct rts_frame *rf = (struct rts_frame*)p->access(hdr_mac::offset_);
	
	assert(pktTx_);
	assert(pktRTS_ == 0);

	/*
	 *  If the size of the packet is larger than the
	 *  RTSThreshold, then perform the RTS/CTS exchange.
	 *
	 *  XXX: also skip if destination is a broadcast
	 */
	if( (u_int32_t) HDR_CMN(pktTx_)->size() < macmib_->RTSThreshold ||
	    (u_int32_t) dst == MAC_BROADCAST) {
		Packet::free(p);
		//p = 0;
		return;
	}

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->size() = ETHER_RTS_LEN;
	ch->iface() = -2;
	ch->error() = 0;

	bzero(rf, MAC_HDR_LEN);

	rf->rf_fc.fc_protocol_version = MAC_ProtocolVersion;
 	rf->rf_fc.fc_type	= MAC_Type_Control;
 	rf->rf_fc.fc_subtype	= MAC_Subtype_RTS;
 	rf->rf_fc.fc_to_ds	= 0;
 	rf->rf_fc.fc_from_ds	= 0;
 	rf->rf_fc.fc_more_frag	= 0;
 	rf->rf_fc.fc_retry	= 0;
 	rf->rf_fc.fc_pwr_mgt	= 0;
 	rf->rf_fc.fc_more_data	= 0;
 	rf->rf_fc.fc_wep	= 0;
 	rf->rf_fc.fc_order	= 0;


	rf->rf_duration = usec(sifs_ + CTS_Time + 2*sifs_  );
	//ETHER_ADDR(rf->rf_ra) = dst;
	STORE4BYTE(&dst, (rf->rf_ra));

	//ETHER_ADDR(rf->rf_ta) = index_;
	STORE4BYTE(&index_, (rf->rf_ta));

	rf->rf_attempt = 0;


	pktRTS_ = p;
	
}

void
Mac802_11mb::sendCTS(int dst, double rts_duration)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct cts_frame *cf = (struct cts_frame*)p->access(hdr_mac::offset_);

	assert(pktCTRL_ == 0);

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->size() = ETHER_CTS_LEN;
	ch->iface() = -2;
	ch->error() = 0;
	//ch->direction() = hdr_cmn::DOWN;
	bzero(cf, MAC_HDR_LEN);

	cf->cf_fc.fc_protocol_version = MAC_ProtocolVersion;
	cf->cf_fc.fc_type	= MAC_Type_Control;
	cf->cf_fc.fc_subtype	= MAC_Subtype_CTS;
 	cf->cf_fc.fc_to_ds	= 0;
 	cf->cf_fc.fc_from_ds	= 0;
 	cf->cf_fc.fc_more_frag	= 0;
 	cf->cf_fc.fc_retry	= 0;
 	cf->cf_fc.fc_pwr_mgt	= 0;
 	cf->cf_fc.fc_more_data	= 0;
 	cf->cf_fc.fc_wep	= 0;
 	cf->cf_fc.fc_order	= 0;
	
	cf->cf_duration = CTS_DURATION(rts_duration); 
 
	// Pradeep Kyasanur: The backoff value to be sent is added here 
	if(backoffSender) 
		{ 
			cf->cf_backoff = backoffSender->getBackoff(dst); 
			cf->cf_src = index_; 
			// printf("Transmission backoff to %d set to %d at %f\n",dst,cf->cf_backoff,(Scheduler::instance()).clock()); 
		} 
	else 
		{ 
			//printf("Sending a CTS packet to %d from %d\n",dst,index_); 
			cf->cf_backoff = 0; 
			cf->cf_src = index_; 
		} 
 
	//Pradeep Kyasanur: If monitor is set and we are using backoff sender, store the backoff value 

	if(monitor_ && backoffSender)       
		monitor_->updateBackoff(dst,cf->cf_backoff); 

	//ETHER_ADDR(cf->cf_ra) = dst;
	STORE4BYTE(&dst, (cf->cf_ra));
	//STORE4BYTE(&dst, (mh->dh_da));

	pktCTRL_ = p;
	
}

void
Mac802_11mb::sendACK(int dst)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct ack_frame *af = (struct ack_frame*)p->access(hdr_mac::offset_);

	assert(pktCTRL_ == 0);

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->size() = ETHER_ACK_LEN;
	ch->iface() = -2;
	ch->error() = 0;
	
	bzero(af, MAC_HDR_LEN);

	af->af_fc.fc_protocol_version = MAC_ProtocolVersion;
 	af->af_fc.fc_type	= MAC_Type_Control;
 	af->af_fc.fc_subtype	= MAC_Subtype_ACK;
 	af->af_fc.fc_to_ds	= 0;
 	af->af_fc.fc_from_ds	= 0;
 	af->af_fc.fc_more_frag	= 0;
 	af->af_fc.fc_retry	= 0;
 	af->af_fc.fc_pwr_mgt	= 0;
 	af->af_fc.fc_more_data	= 0;
 	af->af_fc.fc_wep	= 0;
 	af->af_fc.fc_order	= 0;

	af->af_duration = ACK_DURATION();
	//ETHER_ADDR(af->af_ra) = dst;
	STORE4BYTE(&dst, (af->af_ra));

	pktCTRL_ = p; 
	
}

void
Mac802_11mb::sendDATA(Packet *p)
{
	hdr_cmn* ch = HDR_CMN(p);
	struct hdr_mac802_11* dh = HDR_MAC802_11(p);

	assert(pktTx_ == 0);

	/*
	 * Update the MAC header
	 */
	ch->size() += ETHER_HDR_LEN;

	dh->dh_fc.fc_protocol_version = MAC_ProtocolVersion;
	dh->dh_fc.fc_type       = MAC_Type_Data;
	dh->dh_fc.fc_subtype    = MAC_Subtype_Data;
	//printf(".....p = %x, mac-subtype-%d\n",p,dh->dh_fc.fc_subtype);
	
	dh->dh_fc.fc_to_ds      = 0;
	dh->dh_fc.fc_from_ds    = 0;
	dh->dh_fc.fc_more_frag  = 0;
	dh->dh_fc.fc_retry      = 0;
	dh->dh_fc.fc_pwr_mgt    = 0;
	dh->dh_fc.fc_more_data  = 0;
	dh->dh_fc.fc_wep        = 0;
	dh->dh_fc.fc_order      = 0;



	if((u_int32_t)ETHER_ADDR(dh->dh_da) != MAC_BROADCAST)
		dh->dh_duration =usec(TX_Time(p) + sifs_ + ACK_Time);
	else
		dh->dh_duration = 0;
	pktTx_ = p;
}

/* ======================================================================
   Retransmission Routines
   ====================================================================== */
void
Mac802_11mb::RetransmitRTS()
{
	assert(pktTx_);
	assert(pktRTS_);
	assert(mhBackoff_.busy() == 0);

	macmib_->RTSFailureCount++;

	ssrc_ += 1;			// STA Short Retry Count

	if(ssrc_ >= macmib_->ShortRetryLimit) {
		discard(pktRTS_, DROP_MAC_RETRY_COUNT_EXCEEDED); pktRTS_ = 0;
		/* tell the callback the send operation failed 
		   before discarding the packet */
		hdr_cmn *ch = HDR_CMN(pktTx_);
		if (ch->xmit_failure_) {
                        /*
                         *  Need to remove the MAC header so that 
                         *  re-cycled packets don't keep getting
                         *  bigger.
                         */
                        ch->size() -= ETHER_HDR_LEN;
                        ch->xmit_reason_ = XMIT_REASON_RTS;
                        ch->xmit_failure_(pktTx_->copy(),
                                          ch->xmit_failure_data_);
                }
		//printf("(%d)....discarding RTS:%x\n",index_,pktRTS_);
		discard(pktTx_, DROP_MAC_RETRY_COUNT_EXCEEDED); pktTx_ = 0;
		ssrc_ = 0;
		rst_cw();
	} else {
		//printf("(%d)...retxing RTS:%x\n",index_,pktRTS_);
		struct rts_frame *rf;
		rf = (struct rts_frame*)pktRTS_->access(hdr_mac::offset_);
		rf->rf_fc.fc_retry = 1;

		inc_cw(); 
		
		// Pradeep Kyasanur
		int backoff; 
		struct hdr_mac802_11 *mh = HDR_MAC802_11(pktRTS_); 
		if(backoffReceiver) 
			backoff= backoffReceiver->getBackoff((u_int32_t)ETHER_ADDR(mh->dh_da)); 
		else 
			backoff = (Random::random() % cw_) + 1; 
		mhBackoff_.start(backoff, is_idle()); 

	}
} 

void
Mac802_11mb::RetransmitDATA()
{
	struct hdr_cmn *ch;
	struct hdr_mac802_11 *mh;
	u_int32_t *rcount, *thresh;

	assert(mhBackoff_.busy() == 0);

	assert(pktTx_);
	assert(pktRTS_ == 0);

	ch = HDR_CMN(pktTx_);
	mh = HDR_MAC802_11(pktTx_);

	/*
	 *  Broadcast packets don't get ACKed and therefore
	 *  are never retransmitted.
	 */
	if((u_int32_t)ETHER_ADDR(mh->dh_da) == MAC_BROADCAST) {
		Packet::free(pktTx_); pktTx_ = 0;

		/*
		 * Backoff at end of TX.
		 */
		rst_cw(); 
 
		return;
	}

	macmib_->ACKFailureCount++;

	if((u_int32_t) ch->size() <= macmib_->RTSThreshold) {
		rcount = &ssrc_;
		thresh = &macmib_->ShortRetryLimit;
	}
	else {
		rcount = &slrc_;
		thresh = &macmib_->LongRetryLimit;
	}

	(*rcount)++;

	if(*rcount > *thresh) {
		macmib_->FailedCount++;
		/* tell the callback the send operation failed 
		   before discarding the packet */
		hdr_cmn *ch = HDR_CMN(pktTx_);
		if (ch->xmit_failure_) {
                        ch->size() -= ETHER_HDR_LEN;
                        ch->xmit_reason_ = XMIT_REASON_ACK;
                        ch->xmit_failure_(pktTx_->copy(),
                                          ch->xmit_failure_data_);
                }

		discard(pktTx_, DROP_MAC_RETRY_COUNT_EXCEEDED); pktTx_ = 0;
		//printf("(%d)DATA discarded: count exceeded\n",index_);
		*rcount = 0;
		rst_cw();
	}
	else {
		struct hdr_mac802_11 *dh;
		dh = HDR_MAC802_11(pktTx_);
		dh->dh_fc.fc_retry = 1;

		sendRTS(ETHER_ADDR(mh->dh_da));
		//printf("(%d)retxing data:%x..sendRTS..\n",index_,pktTx_);
		inc_cw(); 
		
		// Pradeep Kyasanur
		int backoff; 
		if(backoffReceiver) 
			backoff= backoffReceiver->getBackoff((u_int32_t)ETHER_ADDR(mh->dh_da)); 
		else 
			backoff = (Random::random() % cw_) + 1; 
		mhBackoff_.start(backoff, is_idle()); 

	}
}

/* ======================================================================
   Incoming Packet Routines
   ====================================================================== */
void
Mac802_11mb::send(Packet *p, Handler *h)
{
	struct hdr_mac802_11* dh = HDR_MAC802_11(p);

	/* 
	 * drop the packet if the node is in sleep mode
	 XXX sleep mode can't stop node from sending packets
	*/
	EnergyModel *em = netif_->node()->energy_model();
	if (em && em->sleep()) {
		em->set_node_sleep(0);
		em->set_node_state(EnergyModel::INROUTE);
	}
	
	callback_ = h;
	sendDATA(p);
	sendRTS(ETHER_ADDR(dh->dh_da));

	/*
	 * Assign the data packet a sequence number.
	 */
	dh->dh_scontrol = sta_seqno_++;

	/*
	 *  If the medium is IDLE, we must wait for a DIFS
	 *  Space before transmitting.
	 */
	if(mhBackoff_.busy() == 0) { // changing the protocol to require counting down IDLE slots always
		//		if(is_idle()) {
		/*
		 * If we are already deferring, there is no
		 * need to reset the Defer timer.
		 */
		//			if(mhDefer_.busy() == 0)
		//				mhDefer_.start(difs_);
		//		}
		/*
		 * If the medium is NOT IDLE, then we start
		 * the backoff timer.
		 */
		//		else { 
		
		rst_cw();

		// Pradeep Kyasanur
		if(backoffReceiver)
			backoffReceiver->resetAttemptNumber(ETHER_ADDR(dh->dh_da));
		int backoff; 
		
		if(backoffReceiver) 
			backoff= backoffReceiver->getBackoff((u_int32_t)ETHER_ADDR(dh->dh_da)); 
		else  
			backoff = (Random::random() % cw_) + 1; 
		mhBackoff_.start(backoff, is_idle()); 

		//		mhBackoff_.start(cw_, is_idle());
		//		}
	}
}

void
Mac802_11mb::recv(Packet *p, Handler *h)
{
	struct hdr_cmn *hdr = HDR_CMN(p);
	/*
	 * Sanity Check
	 */
	assert(initialized());

	/*
	 *  Handle outgoing packets.
	 */
	if(hdr->direction() == hdr_cmn::DOWN) {
                send(p, h);
                return;
        }
	/*
	 *  Handle incoming packets.
	 *
	 *  We just received the 1st bit of a packet on the network
	 *  interface.
	 *
	 */

	/*
	 *  If the interface is currently in transmit mode, then
	 *  it probably won't even see this packet.  However, the
	 *  "air" around me is BUSY so I need to let the packet
	 *  proceed.  Just set the error flag in the common header
	 *  to that the packet gets thrown away.
	 */
	if(tx_active_ && hdr->error() == 0) {
		hdr->error() = 1;
	}

	if(rx_state_ == MAC_IDLE) {
		SET_RX_STATE(MAC_RECV);
		pktRx_ = p;

		/*
		 * Schedule the reception of this packet, in
		 * txtime seconds.
		 */
		mhRecv_.start(TX_Time(p));
	} else {
		/*
		 *  If the power of the incoming packet is smaller than the
		 *  power of the packet currently being received by at least
                 *  the capture threshold, then we ignore the new packet.
		 */
		if(pktRx_->txinfo_.RxPr / p->txinfo_.RxPr >= p->txinfo_.CPThresh) {
			capture(p);
			//			printf("CAPTURE %d at %f\n",index_,(Scheduler::instance()).clock()); 
		} else {
			//			printf("COLL %d at %f\n",index_,(Scheduler::instance()).clock()); 
			collision(p);
		}
	}
}

void
Mac802_11mb::recv_timer()
{
	u_int32_t src;  
	hdr_cmn *ch = HDR_CMN(pktRx_); 
	hdr_mac802_11 *mh = HDR_MAC802_11(pktRx_); 
	u_int32_t dst = ETHER_ADDR(mh->dh_da); 
	// XXX debug 
	//struct cts_frame *cf = (struct cts_frame*)pktRx_->access(hdr_mac::offset_); 
	//u_int32_t src = ETHER_ADDR(mh->dh_sa); 
	 
	u_int8_t  type = mh->dh_fc.fc_type; 
	u_int8_t  subtype = mh->dh_fc.fc_subtype;

	assert(pktRx_);
	assert(rx_state_ == MAC_RECV || rx_state_ == MAC_COLL);
	
        /*
         *  If the interface is in TRANSMIT mode when this packet
         *  "arrives", then I would never have seen it and should
         *  do a silent discard without adjusting the NAV.
         */
        if(tx_active_) {
                Packet::free(pktRx_);
		//		printf("COL-TRANSMIT %d from %d at %f\n",index_,dst,(Scheduler::instance()).clock()); 
                goto done;
        }

	/*
	 * Handle collisions.
	 */
	if(rx_state_ == MAC_COLL) {
		discard(pktRx_, DROP_MAC_COLLISION);
		//		printf("COL-TIMER %d from %d at %f\n",index_,dst,(Scheduler::instance()).clock()); 
		set_nav(usec(eifs_));
		goto done;
	}

	/*
	 * Check to see if this packet was received with enough
	 * bit errors that the current level of FEC still could not
	 * fix all of the problems - ie; after FEC, the checksum still
	 * failed.
	 */
	if( ch->error() ) {
		Packet::free(pktRx_);
		//		printf("ERROR %d from %d at %f\n",index_,dst,(Scheduler::instance()).clock()); 
		set_nav(usec(eifs_));
		goto done;
	}

	/*
	 * IEEE 802.11 specs, section 9.2.5.6
	 *	- update the NAV (Network Allocation Vector)
	 */
	if(dst != (u_int32_t)index_) {
		/* Pradeep: 3/12/03 NAV set by this variable is all that is of interest to us for a scheme
		 *           that counts only the duration fields in RTS/CTS/DATA/ACK to be busy indications
		 *          In addition, if we are receiving a packet above RX_threshold, then during that 
		 *          interval, if NAV is not set, we will have to count the channel to be busy
		 */
		set_nav(mh->dh_duration);

		// Pradeep Kyasanur: Need to set this variable for modified backoff scheme proposed in
		// journal paper
		set_modified_nav(mh->dh_duration);
		
		
	}

        /* tap out - */
        if (tap_ && type == MAC_Type_Data &&
            MAC_Subtype_Data == subtype ) 
		tap_->tap(pktRx_); 
 

	//CODE TO MONTIOR AND MARK THE LAST SEEN PACKET FROM A NODE - added on Jul 31 2002 by Pradeep Kyasanur
	
	if(dst != (u_int32_t)index_ && monitor_)
		{
			src = ETHER_ADDR(mh->dh_sa);
			monitor_->updateLastCount(src);
		}
		 

	/*
	 * Adaptive Fidelity Algorithm Support - neighborhood infomation 
	 * collection
	 *
	 * Hacking: Before filter the packet, log the neighbor node
	 * I can hear the packet, the src is my neighbor
	 */
	if (netif_->node()->energy_model() && 
	    netif_->node()->energy_model()->adaptivefidelity()) {
		src = ETHER_ADDR(mh->dh_sa);
		netif_->node()->energy_model()->add_neighbor(src);
	}
	/*
	 * Address Filtering
	 */
	if(dst != (u_int32_t)index_ && dst != MAC_BROADCAST) {
		/*
		 *  We don't want to log this event, so we just free
		 *  the packet instead of calling the drop routine.
		 */
		discard(pktRx_, "---");
		goto done;
	}

	switch(type) {

	case MAC_Type_Management:
		discard(pktRx_, DROP_MAC_PACKET_ERROR);
		goto done;
		break;

	case MAC_Type_Control:
		switch(subtype) {
		case MAC_Subtype_RTS:
			recvRTS(pktRx_);
			break;
		case MAC_Subtype_CTS:
			recvCTS(pktRx_);
			break;
		case MAC_Subtype_ACK:
			recvACK(pktRx_);
			break;
		default:
			fprintf(stderr,"recvTimer1:Invalid MAC Control Subtype %x\n",
				subtype);
			exit(1);
		}
		break;
	case MAC_Type_Data:
		switch(subtype) {
		case MAC_Subtype_Data:
			recvDATA(pktRx_);
			break;
		default:
			fprintf(stderr, "recv_timer2:Invalid MAC Data Subtype %x\n",
				subtype);
			exit(1);
		}
		break;
	default:
		fprintf(stderr, "recv_timer3:Invalid MAC Type %x\n", subtype);
		exit(1);
	}
 done:
	pktRx_ = 0;
	rx_resume();
}


void
Mac802_11mb::recvRTS(Packet *p)
{
	struct rts_frame *rf = (struct rts_frame*)p->access(hdr_mac::offset_);

	if(tx_state_ != MAC_IDLE) {
		discard(p, DROP_MAC_BUSY);
		return;
	}

	/*
	 *  If I'm responding to someone else, discard this RTS.
	 */
	if(pktCTRL_) {
		discard(p, DROP_MAC_BUSY);
		return;
	} 

	/* pass the packet to monitor */ 
	if(monitor_) 
		monitor_->recv(pktRx_); 
 

	sendCTS(ETHER_ADDR(rf->rf_ta), rf->rf_duration);

	//	printf("Receiving a RTS packet in %d at %f\n",index_,(Scheduler::instance()).clock()); 

	/*
	 *  Stop deferring - will be reset in tx_resume().
	 */
	if(mhDefer_.busy()) mhDefer_.stop();

	tx_resume();

	mac_log(p);
}


void
Mac802_11mb::recvCTS(Packet *p)
{
	if(tx_state_ != MAC_RTS) {
		discard(p, DROP_MAC_INVALID_STATE);
		return;
	}

	assert(pktRTS_);
	Packet::free(pktRTS_); pktRTS_ = 0;

	assert(pktTx_);
	// debug
	//struct hdr_mac802_11mb *mh = HDR_MAC802_11mb(pktTx_);
	//printf("(%d):recvCTS:pktTx_-%x,mac-subtype-%d & pktCTS_:%x\n",index_,pktTx_,mh->dh_fc.fc_subtype,p);
	
	mhSend_.stop();

	/*
	 * The successful reception of this CTS packet implies
	 * that our RTS was successful.  Hence, we can reset
	 * the Short Retry Count and the CW.
	 */
	ssrc_ = 0;
	rst_cw();

	tx_resume();
 
	/* set the backoff value if needed */ 
	struct cts_frame *cf = (struct cts_frame*)p->access(hdr_mac::offset_); 
	if(backoffReceiver) 
		{ 
			//	printf("At node %d from node %d ",index_,cf->cf_src); 
			backoffReceiver->setBackoff(p); 
		} 
 
	//	printf("CTS received %d at %f\n",index_,(Scheduler::instance()).clock()); 

	mac_log(p); 
 
	
}

void
Mac802_11mb::recvDATA(Packet *p)
{
	struct hdr_mac802_11 *dh = HDR_MAC802_11(p);
	u_int32_t dst, src, size;

	{	struct hdr_cmn *ch = HDR_CMN(p);

	dst = ETHER_ADDR(dh->dh_da);
	src = ETHER_ADDR(dh->dh_sa);
	size = ch->size();

	/*
	 * Adjust the MAC packet size - ie; strip
	 * off the mac header
	 */
	ch->size() -= ETHER_HDR_LEN;
	ch->num_forwards() += 1;
	}

	/*
	 *  If we sent a CTS, clean up...
	 */
	if(dst != MAC_BROADCAST) {
		if(size >= macmib_->RTSThreshold) {
			if (tx_state_ == MAC_CTS) {
				assert(pktCTRL_);
				Packet::free(pktCTRL_); pktCTRL_ = 0;
				mhSend_.stop();
				/*
				 * Our CTS got through.
				 */
				//printf("(%d): RECVING DATA!\n",index_);

				// Pradeep Kyasanur: 
				// THIS IS A BUG - From discussions on Manet on Feb 4 2003
				// The ssrc_ should not be set to 0 because we may have been
				// sending RTS to some other node 

				//				ssrc_ = 0;
				//				rst_cw();
			}
			else {
				discard(p, DROP_MAC_BUSY);
				// In this case, have to restore the backoff value in the Monitor 
				if(monitor_) 
					monitor_->restoreBackoff(dst); 
				//printf("(%d)..discard DATA\n",index_);
				return;
			}
			sendACK(src);
			tx_resume();
		}
		/*
		 *  We did not send a CTS and there's no
		 *  room to buffer an ACK.
		 */
		else {
			if(pktCTRL_) {
				discard(p, DROP_MAC_BUSY);
				return;
			}
			sendACK(src);
			if(mhSend_.busy() == 0)
				tx_resume();
		}
	}

	/* ============================================================
	   Make/update an entry in our sequence number cache.
	   ============================================================ */

	/* Changed by Debojyoti Dutta. This upper loop of if{}else was 
	   suggested by Joerg Diederich <dieder@ibr.cs.tu-bs.de>. 
	   Changed on 19th Oct'2000 */

        if(dst != MAC_BROADCAST) {
                if (src < (u_int32_t) cache_node_count_) {
                        Host *h = &cache_[src];

                        if(h->seqno && h->seqno == dh->dh_scontrol) {
                                discard(p, DROP_MAC_DUPLICATE);
                                return;
                        }
                        h->seqno = dh->dh_scontrol;
                } else {
			static int count = 0;
			if (++count <= 10) {
				//printf ("MAC_802_11: accessing MAC cache_ array out of range (src %u, dst %u, size %d)!\n", src, dst, cache_node_count_);
				//if (count == 10)
				//printf ("[suppressing additional MAC cache_ warnings]\n");
			};
		};
	}

	/*
	 *  Pass the packet up to the link-layer.
	 *  XXX - we could schedule an event to account
	 *  for this processing delay.
	 */
	p->incoming = 1;
	uptarget_->recv(p, (Handler*) 0);
}


void
Mac802_11mb::recvACK(Packet *p)
{
	
	struct hdr_cmn *ch = HDR_CMN(p);

	if(tx_state_ != MAC_SEND) {
		discard(p, DROP_MAC_INVALID_STATE);
		return;
	}
	//	printf("(%d)...................recving ACK:%x\n",index_,p);
	assert(pktTx_);
	Packet::free(pktTx_); pktTx_ = 0;

	mhSend_.stop();

	/*
	 * The successful reception of this ACK packet implies
	 * that our DATA transmission was successful.  Hence,
	 * we can reset the Short/Long Retry Count and the CW.
	 */
	if((u_int32_t) ch->size() <= macmib_->RTSThreshold)
		ssrc_ = 0;
	else
		slrc_ = 0;

	/*
	 * Backoff before sending again.
	 */
	rst_cw();
	assert(mhBackoff_.busy() == 0); 
 
 
	/* Pradeep Kyasanur: I need to start this only if there is an RTS waiting to be sent */ 

	if(pktRTS_)
		{

	
			int backoff;
			struct hdr_mac802_11 *mh = HDR_MAC802_11(pktRTS_);
			if(backoffReceiver)
				backoffReceiver->resetAttemptNumber(ETHER_ADDR(mh->dh_da));
			if(backoffReceiver)
				backoff= backoffReceiver->getBackoff((u_int32_t)ETHER_ADDR(mh->dh_da));
			else 
				backoff = (Random::random() % cw_) + 1;
			mhBackoff_.start(backoff, is_idle());  // THIS IS A PROBLEM
	

			//	mhBackoff_.start(cw_, is_idle());

		}
	tx_resume();

	mac_log(p);
}


// -----------------------------------------------------------------------------------------------
/* BELOW IS NEW CODE ADDED BY PRADEEP KYASANUR */


void
Mac802_11mb ::trace(char* fmt, ...)
{
	va_list ap;
        
	if (!misbtarget_) return;
	va_start(ap, fmt);
	vsprintf(misbtarget_->pt_->buffer(), fmt, ap);
	misbtarget_->pt_->dump();
	va_end(ap);
}


/* functions for Monitoring the traffic */
void Monitor::recv(const Packet *p)  {  // capture any type of packet received from MAC to monitor the traffic


	if(p == NULL)
		fprintf(stderr,"NULL PACKET !!! \n");
	hdr_mac802_11 *mh = HDR_MAC802_11(p);
	u_int32_t dst = ETHER_ADDR(mh->dh_sa);
	
	u_int8_t  type = mh->dh_fc.fc_type;
	u_int8_t  subtype = mh->dh_fc.fc_subtype;
	

        
	/* currently assume that we are checking only for RTS packets */

	struct rts_frame *rf = (struct rts_frame*)p->access(hdr_mac::offset_);

	u_int16_t attemptNum=rf->rf_attempt;


	// use idle timer to compute success probability
	int curCollCount = idleCounter->getCurrentCount(); 

	
	int diffCount = attemptNum;

	if(diffCount!=0) {
		averageAttempt = alpha * averageAttempt + (1- alpha) * diffCount;
		prevCollCount=curCollCount; }


	// seeing a packet for the firstTime from this node
	if(nodeTable.find(dst)==nodeTable.end() ) 
		nodeTable[dst]=new NodeInfo(dst,THRESH);
	else	{
		int curIdleSlot=idleCounter->getCurrentCount();

		// this is the slots waited by sender between transmissions as seen by me
		int actualSlotDiff = curIdleSlot - nodeTable[dst]->lastCount; 
		int startCW=31;
		int expectedSlotDiff=nodeTable[dst]->givenBackoff;
		int startVal=expectedSlotDiff;
		int temp=attemptNum;

		// this is used later to compute penalty also
		int attemptCount = 1;
		while(attemptNum > 1)		{
			int tempVal;

			if(startCW < 1023)
				startCW=startCW*2+1;
				
			tempVal = ( ( ( ( startVal + dst ) % 32 ) * 5 + (2 * attemptCount + 1 ) ) % 32 ) * (startCW / 32) + 1;

			attemptCount ++;	
			expectedSlotDiff+=tempVal;
			attemptNum--;
		}
		attemptNum = temp;



		int allowedVariance = (int)(expectedSlotDiff * (1.0 - jitter_alpha)) ;
	

		// Log the packet
		mac->trace("PRD %f %d %d %d\n",(Scheduler::instance()).clock(),dst,actualSlotDiff,expectedSlotDiff);
	
		// Handling the window
		
		if(nodeTable[dst]->windowFull)	{
			nodeTable[dst]->miscount-=nodeTable[dst]->window[nodeTable[dst]->windowIndex];
			nodeTable[dst]->window[nodeTable[dst]->windowIndex]=0;
		}

		nodeTable[dst]->miscount+=(expectedSlotDiff-actualSlotDiff);
		nodeTable[dst]->window[nodeTable[dst]->windowIndex]=(expectedSlotDiff-actualSlotDiff);

		if( (actualSlotDiff < expectedSlotDiff) && ((expectedSlotDiff - actualSlotDiff) > allowedVariance ))	{
			
			nodeTable[dst]->remainingBackoff+=(expectedSlotDiff-actualSlotDiff) - allowedVariance; // note the value to be added for future attempts
		
			double idleProb = idleCounter->getCollisionProbability();
			double extraSlot = 0;
		
			// CODE FOR COMPUTING PENALTY

			averageBackoff  = alpha * averageBackoff  + (1- alpha) * expectedSlotDiff;
			double beta = (double) actualSlotDiff / expectedSlotDiff;
			double p = averageAttempt / (averageBackoff + averageAttempt);
			double Pidle = 1.0 / idleProb;

			extraSlot = p * ( 1 - beta) / Pidle;
		
			if( extraSlot - (int)extraSlot > 0.5)
				extraSlot +=1;
			attemptNum = (int) extraSlot;
		
			// COMPUTING PENALTY AS BEFORE. attemptCount, startCW and startVal are used from the earlier calculation of expectedSlotDiff
			while(attemptNum >= 1) 	{
				if(startCW < 1023)
					startCW=startCW*2+1;
				
				//COMMENT NEXT LINE FOR ADDITIONAL PENALTY NOT TO BE ADDED. Details of 
				// why additional penalty has to be added is in technical paper of Aug 2002
				nodeTable[dst]->remainingBackoff+= ( ( ( ( startVal + dst ) % 32 ) * 5 + (2 * attemptCount + 1 ) ) % 32 ) * (startCW / 32) + 1;
						
				attemptCount ++;	 
				attemptNum--;
			}
				

				
			if(nodeTable[dst]->miscount >= K)
				mac->trace("MBD %f %d %d\n",(Scheduler::instance()).clock(),dst,nodeTable[dst]->remainingBackoff);
		}
				
		if(++nodeTable[dst]->windowIndex == THRESH)				{
			nodeTable[dst]->windowIndex=0;
			nodeTable[dst]->windowFull=1;
		}
				
		nodeTable[dst]->lastCount=idleCounter->getCurrentCount();
	}



}


void Monitor :: updateLastCount(const Packet *p)
{

	if(p == NULL)
		fprintf(stderr,"NULL PACKET !!! \n");
	hdr_mac802_11 *mh = HDR_MAC802_11(p);
	u_int32_t dst = ETHER_ADDR(mh->dh_da);
	if(nodeTable.find(dst)==nodeTable.end() ) // seeing a packet for the firstTime from this node
		nodeTable[dst]=new NodeInfo(dst,THRESH);
	nodeTable[dst]->lastCount=idleCounter->getCurrentCount();


}

void Monitor :: updateLastCount(u_int32_t dst)
{
	if(nodeTable.find(dst)==nodeTable.end() ) // seeing a packet for the firstTime from this node
		nodeTable[dst]=new NodeInfo(dst,THRESH);
	nodeTable[dst]->lastCount=idleCounter->getCurrentCount();

}


void Monitor::updateBackoff(u_int32_t dst,int backoff)
{
	if(nodeTable.find(dst)==nodeTable.end() ) // seeing a packet for the firstTime from this node
		nodeTable[dst]=new NodeInfo(dst,THRESH);
	nodeTable[dst]->oldBackoff= nodeTable[dst]->givenBackoff;
	nodeTable[dst]->givenBackoff=backoff;
	

}

// to restore the backoff when CTS containing next backoff to use is lost 
void Monitor :: restoreBackoff(const Packet *p)
{
	hdr_mac802_11 *mh = HDR_MAC802_11(p);
	u_int32_t dst = ETHER_ADDR(mh->dh_da);
	if(nodeTable.find(dst)==nodeTable.end() ) // This should not happen, but if it does then code will not break
		nodeTable[dst]=new NodeInfo(dst,THRESH);
	nodeTable[dst]->givenBackoff=nodeTable[dst]->oldBackoff;

}

// to restore the backoff when CTS containing next backoff to use is lost 
void Monitor :: restoreBackoff(u_int32_t dst)
{

	if(nodeTable.find(dst)==nodeTable.end() ) // This should not happen, but if it does then code will not break
		nodeTable[dst]=new NodeInfo(dst,THRESH);
	nodeTable[dst]->givenBackoff=nodeTable[dst]->oldBackoff;

}


/* Code for BackoffSender class which calculates the next backoff values to be used by other
   nodes to transmit to this node */

// get the next backoff value for the destination
u_int16_t BackoffSender :: getBackoff(u_int32_t dst) 
{
	float PENALTY = 1.0;
	int increment=0;
	if(mon)
		{
			increment=(int)(mon->nodeTable[dst]->remainingBackoff*PENALTY);	
			mon->nodeTable[dst]->remainingBackoff=0;
		}
	int backoff = ( Random :: random() % CW );


	return backoff + 1+increment ; // to ensure that it is never zero
}




/* KY: 05/18/03: To model a misbehaving node  with adaptive misbehavior
 * 1. The aim of the proposed modifications is to allow the misbehaving node to suitably adjust magnitude of misbehavior
 *    based on the assigned backoffs
 * 2. The misbehaving node decides a penalty has been assigned if the assigned backoff is greater than CW_min = 31. In that
 *    case, the misbehaving node waits for the assigned backoff and decrements the magnitude of misbehavior
 * 3. This behavior mimics a node which trys to adaptively misbehave without getting caught
 */

// currently it does not store information about which destination has to use this value	
void BackoffReceiver:: setBackoff(Packet *p)
{

	if(p == NULL)
		fprintf(stderr,"NULL PACKET !!! \n");

	struct cts_frame *cf = (struct cts_frame*)p->access(hdr_mac::offset_);
	u_int32_t dst = cf->cf_src;
	u_int32_t value = cf->cf_backoff;

	
	if(value == 0) // node is not specifying a value of backoff to use so set it to some random value
		nodeTable[dst]=(Random::random() % mac->getCW()) + 1;
	else { 	

		// KY: 05/19/03 - Code for adaptive misbehavior 
		if(dec > 0 ) { // only if adaptive misbehavior is enabled		
			
			if(value > 31 && misb >= 0 ) { // This code is triggered only when assigned backoff is greater than CW_min implying a penalty and adaptive misbehavior has been desired by setting misb to >= 0
				misb = misb + (100 - misb) / 2; // reduce misbehavior by half
				backoffTimer->set_misbehavior_percentage(misb);
		
			}
			else if(misb > dec) {
				misb -= dec; // else increase misbehavior by dec%
				backoffTimer->set_misbehavior_percentage(misb);
			}

		}
		nodeTable[dst]=value;
		attemptNum[dst]=0;
	}

}


int BackoffReceiver :: getBackoff(u_int32_t dst)
{

	if(nodeTable.find(dst)==nodeTable.end() ) // seeing request for this node for the first time
		{
			nodeTable[dst]=(Random::random() % mac->getCW()) + 1;
			attemptNum[dst]=0;
		}
	int val ;
	if(attemptNum[dst]==0) 
		val= nodeTable[dst];
	else
		val = curTable[dst];

	// The function for recomputing backoff values for subsequent attempts is just  (previous value )*nodeID %CW
	int CW = mac->getCW();



	if( CW < 1023)
		CW=CW*2+1;

	attemptNum[dst]++;

	/* FUNCTION f used to generate the backoff value after collision. Discussion on the properties
	   of this function are in technical report of Aug 2002

	f = { [ (origBackoff + nodeID)%CW_min ] * 5 + [2 * attemptNum + 1] } % CW_min

	and dividing by CW_min gives a uniformly distributed integer b/w 0 and 1. Scaling
	this by CW gives the  required backoff value

	CW_min is assumed to be 32

	*/

	// + 1 is added as I do not want 0 as a value

	curTable[dst]= ( ( ( ( nodeTable[dst] + mac->getNodeId() ) % 32 ) * 5 + (2 * attemptNum[dst] + 1 ) ) % 32 ) * (CW / 32) + 1;

	
	return val;
}

u_int16_t BackoffReceiver :: getAttemptNumber(u_int32_t dst)
{
	if(attemptNum.find(dst)==attemptNum.end())
		return 0;
	
	return attemptNum[dst];


}

void BackoffReceiver :: resetAttemptNumber(u_int32_t dst)
{
	attemptNum[dst] = 0;
}


// -------------------------- End  of code written by Pradeep Kyasanur ----------
